/* Questo sketch disegna un manipolatore planare a due link.
L'organo terminale del robot si posiziona dove si clicca
col mouse (segnalando se il punto è fuori dello spazio di lavoro).
Le frecce destra e sinistra permettono di selezionare la soluzione
gomito alto o gomito basso. Le frecce SU e GIÙ permettono di
modificare la costante Kp (tra un valore minimo e massimo) che 
caratterizza la velocità con cui il robot si muove verso la 
posizione desiderata. */

// Coordinate base robot
float xBase,yBase;
// Coordinate desiderate effettore immesse con click mouse (rispetto base robot)
float xDes = 60;
float yDes = 40;
// lunghezze link robot
float L1 = 100; // lunghezza primo link
float L2 = 70; // lunghezza secondo link  
float[] theta = new float[2];
float argCos2 = 0;
float C2 = 0;
float S2 = 0;
float theta1Des = 0;
float theta2Des = 0;
float kp = .1; // la kp del controllo
float precisione = .001; // la differenza tra angolo finale e desiderato
float gomito = 1;

void setup() {
size(600, 400);
xBase = width/2;
yBase = height/2;
smooth();
}

void draw() 
{
  background(204);
  
  // Per modificare il kp e il gomito alto o basso
  if (keyPressed)
  {
    if (keyCode == UP)
    {
      kp += .01;
      if (kp > .5)
      {
        kp = .5;
      } 
    }
    if (keyCode == DOWN)
    {
      kp -= .01;
      if (kp < 0.01)
      {
        kp = 0.01;
      }
    }
    if (keyCode == LEFT)
    {
      gomito = -1;
    }
    if (keyCode == RIGHT)
    {
      gomito = 1;
    }
  }    
  
  // inserisci coordinate effettore desiderate
  if (mousePressed)
  {
    xDes = mouseX-xBase;
    yDes = yBase-mouseY;
  }

  // calcola coordinate giunto:
  argCos2 = (pow(xDes,2)+pow(yDes,2)-pow(L1,2)-pow(L2,2))/(2*L1*L2);
  if (abs(argCos2)<=1)
  {
    theta2Des = gomito*acos(argCos2);
    C2 = cos(theta2Des);
    S2 = sin(theta2Des);    
    theta1Des = atan2(-L2*S2*xDes+(L1+L2*C2)*yDes,(L1+L2*C2)*xDes+L2*S2*yDes);
    if (abs(theta[0]-theta1Des)>precisione)
    {
      theta[0] += kp*(theta1Des-theta[0]);
    }
    if (abs(theta[1]-theta2Des)>precisione)
    {
      theta[1] += kp*(theta2Des-theta[1]);
    }
  }
  else
  {
    textSize(25);
    fill(255,0,0);
    text("posizione fuori dello spazio di lavoro",width-500,height-20); 
  }
    
  robot(xBase,yBase,theta[0],theta[1]);
  
  textSize(25);
  fill(0,0,256);
  text("kp = ",width-170,30); 
  text(kp,width-110,30);
  if (gomito == -1)
  {
    fill(0,256,0);
    text("Gomito alto",width-170,60); 
  }
  else
  {
    fill(0,256,0);
    text("Gomito basso",width-170,60); 
  }  
  textSize(15);
  fill(0);
  text("xDes = ",10,20); 
  text(xDes,70,20);
  text("yDes = ",150,20); 
  text(yDes,210,20);
  text("xEff = ",10,50); 
  text(L1*cos(theta[0])+L2*cos(theta[0]+theta[1]),70,50);
  text("yEff = ",150,50); 
  text((L1*sin(theta[0])+L2*sin(theta[0]+theta[1])),210,50);
}

void robot(float x, float y, float theta1, float theta2)
{
  // Qui le coordinate si riferiscono allo schermo: (0,0) = in alto a sinistra
  // coordinate grafiche primo giunto
  float x1 = x+L1*cos(theta1);
  float y1 = y-L1*sin(theta1);  // Il meno per la convenzione grafica della y positiva verso il basso
  // coordinate grafiche secondo giunto  
  float x2 = x1+L2*cos(theta1+theta2);
  float y2 = y1-L2*sin(theta1+theta2);  
  
  noStroke();
  fill(0);
  // disegno base (nera)
  rect(x-15,y-15,30,30);
  fill(255,0,0);
  
  // disegno giunti (rossi)
  ellipse(x,y,15,15);
//  fill(255,0,0);
  ellipse(x1,y1,15,15);

  stroke(0);
  strokeWeight(8);
  line(x,y,x1,y1);
  line(x1,y1,x2,y2);
  
  // disegno la pinza
  float xPuntoP1 = x2-7*sin(theta1+theta2);
  float yPuntoP1 = y2-7*cos(theta1+theta2);  
  float xPuntoP2 = x2+7*sin(theta1+theta2);
  float yPuntoP2 = y2+7*cos(theta1+theta2);  
  float xPuntoP11 = xPuntoP1+8*cos(theta1+theta2);
  float yPuntoP11 = yPuntoP1-8*sin(theta1+theta2);
  float xPuntoP22 = xPuntoP2+8*cos(theta1+theta2);
  float yPuntoP22 = yPuntoP2-8*sin(theta1+theta2);
  line(xPuntoP1,yPuntoP1,xPuntoP2,yPuntoP2);
  line(xPuntoP1,yPuntoP1,xPuntoP11,yPuntoP11);
  line(xPuntoP2,yPuntoP2,xPuntoP22,yPuntoP22);
}