/* Questo sketch disegna un manipolatore planare a tre link.
L'organo terminale del robot si posiziona dove si clicca
col mouse (segnalando se il punto è fuori dello spazio di lavoro).
Le frecce destra e sinistra permettono di selezionare 
l'orientamento desiderato della pinza. Le frecce SU e GIÙ permettono di
modificare la costante Kp (tra un valore minimo e massimo) che 
caratterizza la velocità con cui il robot si muove verso la 
posizione desiderata. La soluzione gomito basso o gomito alto viene 
scelta automaticamente per minimizzare lo spostamento dei vari link
del robot: la funzione ottimizzata è quella riportata alle righe 147 e 156
di questo sketch (i coefficienti di costo c1, c2 e c3 sono modificabili da 
tastiera premendo i tasti c, C, x, X, v e V: all'aumentare di c1 rispetto 
a c2 e c3, il robot tenderà a non muovere il primo link). A schermo viene 
riportato il valore delle coordinate xDes, yDes e phiDes desiderate, quello delle 
variabili di giunto theta1, theta2 e theta3 che risolvono la cinematica inversa e 
quello delle variabili effettivamente raggiunte dal robot xEff, yEff e phiEff (molto 
vicine ai valori desiderati se la soluzione è corretta). Sono riportati anche i valori 
dei coefficienti di costo c1, c2 e c3. */


// Coordinate centro robot
float xBase,yBase;
// Coordinate desiderate effettore immesse con click mouse (rispetto base robot)
float xDes = 60; // x pinza desiderato
float yDes = 40; // y pinza desiderato
float phiDes = PI/2; // angolo pinza desiderato
float x3Des; // x giunto 3 desiderato
float y3Des; // y giunto 3 desiderato
// lunghezze link robot
float L1 = 100; // lunghezza primo link
float L2 = 70; // lunghezza secondo link  
float L3 = 30; // lunghezza terzo link  
float[] theta = new float[3];
float argCos2 = 0;
float C2 = 0;
float S2 = 0;
float theta1Des = 0;
float theta2Des = 0;
float theta3Des = 0;
float kp = .1; // la kp del controllo
float precisione = .001; // la differenza tra angolo finale e desiderato
float gomito = 1;
float costoA,costoB;
float c1 = 5;
float c2 = 0.5;
float c3 = 0.1;

void setup() {
size(600, 400);
smooth();
xBase = width/2;
yBase = height/2;
theta[0] = 0;
theta[1] = PI/2;
theta[2] = 0;
}

void draw() 
{
  background(204);
  
  // Per modificare il kp e il gomito alto o basso
  if (keyPressed)
  {
    if (keyCode == UP)
    {
      kp += .001;
      if (kp > .5)
      {
        kp = .5;
      }
    }
    if (keyCode == DOWN)
    {
      kp -= .001;
      if (kp < 0.01)
      {
        kp = 0.01;
      }
    }
    if (keyCode == LEFT)
    {
      phiDes += .02;
    }
    if (keyCode == RIGHT)
    {
      phiDes -= .02;
    }
    if (key == 'C')
    {
      c1 += .1;
    }
    if (key == 'c')
    {
      c1 -= .1;
      if (c1 < 0.1)
      {
        c1 = 0.1;
      }
    }    
    if (key == 'V')
    {
      c2 += .1;
    }
    if (key == 'v')
    {
      c2 -= .1;
      if (c2 < 0.1)
      {
        c2 = 0.1;
      }
    }    
    if (key == 'X')
    {
      c3 += .1;
    }
    if (key == 'x')
    {
      c3 -= .1;
      if (c3 < 0.1)
      {
        c3 = 0.1;
      }
    }    
  }    
  
  // inserisci coordinate effettore desiderate
  if (mousePressed)
  {
    xDes = mouseX-xBase;
    yDes = yBase-mouseY;    
  }

  // calcola coordinate giunto 3:
  x3Des = xDes - L3*cos(phiDes);
  y3Des = yDes - L3*sin(phiDes);  
  argCos2 = (pow(x3Des,2)+pow(y3Des,2)-pow(L1,2)-pow(L2,2))/(2*L1*L2);
  if (abs(argCos2)<=1)
  {
    gomito = 1; // Soluzione gomito BASSO
    theta2Des = gomito*acos(argCos2);
    C2 = cos(theta2Des);
    S2 = sin(theta2Des);    
    theta1Des = atan2(-L2*S2*x3Des+(L1+L2*C2)*y3Des,(L1+L2*C2)*x3Des+L2*S2*y3Des);
    theta3Des = phiDes - theta1Des - theta2Des;
    // Calcolo il costo associato allo spostamento effettuato se scelgo la soluzione gomito BASSO
    costoB = c1*abs(theta1Des-theta[0]) + c2*abs(theta2Des-theta[1]) + c3*abs(theta3Des-theta[2]);

    gomito = -1;  // Soluzione gomito ALTO
    theta2Des = gomito*acos(argCos2);
    C2 = cos(theta2Des);
    S2 = sin(theta2Des);    
    theta1Des = atan2(-L2*S2*x3Des+(L1+L2*C2)*y3Des,(L1+L2*C2)*x3Des+L2*S2*y3Des);
    theta3Des = phiDes - theta1Des - theta2Des;
    // Calcolo il costo associato allo spostamento effettuato se scelgo la soluzione gomito ALTO
    costoA = c1*abs(theta1Des-theta[0]) + c2*abs(theta2Des-theta[1]) + c3*abs(theta3Des-theta[2]);

    if (costoB < costoA) // Scelgo la soluzione che costa di meno
    {
      theta2Des = -theta2Des;
      S2 = - S2;
      theta1Des = atan2(-L2*S2*x3Des+(L1+L2*C2)*y3Des,(L1+L2*C2)*x3Des+L2*S2*y3Des);
      theta3Des = phiDes - theta1Des - theta2Des;
      gomito = 1;
    }
    // Modifico gli angoli del robot verso i loro valori desiderati  
    if (abs(theta[0]-theta1Des)>precisione)
    {
      theta[0] += kp*(theta1Des-theta[0]);
    }
    if (abs(theta[1]-theta2Des)>precisione)
    {
      theta[1] += kp*(theta2Des-theta[1]);
    }
    if (abs(theta[2]-theta3Des)>precisione)
    {
      theta[2] += kp*(theta3Des-theta[2]);
    }
  }
  else
  {
    textSize(25);
    fill(255,0,0);
    text("posizione fuori dello spazio di lavoro",width-500,height-20); 
  }
    
  robot(xBase,yBase,theta[0],theta[1],theta[2]);
  
  textSize(25);
  fill(0);
  text("kp = ",width-150,30); 
  text(kp,width-90,30);
  if (gomito == -1)
  {
    fill(0,255,0);
    text("Gomito alto",width-170,height-40); 
  }
  else
  {
    fill(0,255,0);
    text("Gomito basso",width-170,height-40); 
  }  
  textSize(15);
  fill(0,0,255);
  text("xDes = ",10,20); 
  text(xDes,70,20);
  text("yDes = ",150,20); 
  text(yDes,210,20);
  text("phiDes = ",290,20); 
  text(phiDes*180/PI % 360,360,20);
  fill(255,0,0);
  text("theta1 = ",10,50); 
  text(theta[0]*180/PI % 360,70,50);
  text("theta2 = ",150,50); 
  text(theta[1]*180/PI % 360,210,50);
  text("theta3 = ",290,50); 
  text(theta[2]*180/PI % 360,360,50);  
  fill(255);
  text("xEff = ",10,80); 
  text(L1*cos(theta[0])+L2*cos(theta[0]+theta[1])+L3*cos(theta[0]+theta[1]+theta[2]),70,80);
  text("yEff = ",150,80); 
  text((L1*sin(theta[0])+L2*sin(theta[0]+theta[1]))+L3*sin(theta[0]+theta[1]+theta[2]),210,80);
  text("phiEff = ",290,80); 
  text((theta[0]+theta[1]+theta[2])*180/PI,360,80);
  fill(0);
  text("c1 = ",10,height-80); 
  text(c1,60,height-80);
  text("c2 = ",10,height-50); 
  text(c2,60,height-50);
  text("c3 = ",10,height-20); 
  text(c3,60,height-20);
}

void robot(float x, float y, float theta1, float theta2, float theta3)
{
  // Qui le coordinate si riferiscono allo schermo: (0,0) = in alto a sinistra
  // coordinate grafiche primo giunto
  float x1 = x+L1*cos(theta1); 
  float y1 = y-L1*sin(theta1);  // Il meno per la convenzione grafica della y positiva verso il basso  
  // coordinate grafiche secondo giunto  
  float x2 = x1+L2*cos(theta1+theta2);
  float y2 = y1-L2*sin(theta1+theta2);  
  // coordinate grafiche terzo giunto  
  float x3 = x2+L3*cos(theta1+theta2+theta3);
  float y3 = y2-L3*sin(theta1+theta2+theta3);  
  
  noStroke();
  fill(0);
  // disegno base (nera)
  rect(x-15,y-15,30,30);
  fill(255,0,0);
  
  // disegno giunti (rossi)
  ellipse(x,y,15,15);
//  fill(255,0,0);
  ellipse(x1,y1,15,15);
  ellipse(x2,y2,15,15);

  // disegno link
  stroke(0);
  strokeWeight(8);
  line(x,y,x1,y1);
  line(x1,y1,x2,y2);
  line(x2,y2,x3,y3);  

  // disegno pinza  
  float xPuntoP1 = x3-7*sin(theta1+theta2+theta3);
  float yPuntoP1 = y3-7*cos(theta1+theta2+theta3);
  float xPuntoP2 = x3+7*sin(theta1+theta2+theta3);
  float yPuntoP2 = y3+7*cos(theta1+theta2+theta3);
  float xPuntoP11 = xPuntoP1+8*cos(theta1+theta2+theta3);
  float yPuntoP11 = yPuntoP1-8*sin(theta1+theta2+theta3);
  float xPuntoP22 = xPuntoP2+8*cos(theta1+theta2+theta3);
  float yPuntoP22 = yPuntoP2-8*sin(theta1+theta2+theta3);
  line(xPuntoP1,yPuntoP1,xPuntoP2,yPuntoP2);
  line(xPuntoP1,yPuntoP1,xPuntoP11,yPuntoP11);
  line(xPuntoP2,yPuntoP2,xPuntoP22,yPuntoP22);
}