/* Questo sketch disegna un manipolatore planare a due link
il cui organo terminale si muove in modo circolare con 
velocità angolare omega lungo un cerchio centrato in 
(xCentro,yCentro), di raggio rCerchio.
Le coordinate del centro del cerchio possono essere impostate
cliccando col mouse. Il raggio aumenta o diminuisce coi
tasti '+' e '-'. Le frecce destra e sinistra permettono di 
modificare la velocità angolare omega entro (-0.1,0.1).
Le frecce SU e GIÙ al solito permettono di modificare la
costante Kp (tra un valore minimo e massimo) che caratterizza
la velocità con cui il robot si muove verso la posizione 
desiderata. Alcuni accorgimenti sono stati introdotti per
evitare salti bruschi nel movimento del robot (vedere in
particolare l'aggiunta di nGiri*2*PI alla funzione atan2 che
calcola theta1Des e l'aggiunta di una fase deltaPhi 
nell'argomento di sin(omega*tempo+deltaPhi) e 
cos(omega*tempo+deltaPhi). */


// Coordinate centro robot
float xBase,yBase;

// Variabili per disegnare cerchio
float tempo;
float xCentro,yCentro;
float rCerchio = 100;
float omega = .1; // velocità angolare 
float omegaOld = .1;
float deltaPhi = 0;

// Coordinate desiderate effettore immesse con click mouse (rispetto base robot)
float xDes;
float yDes;
// lunghezze link robot
float L1 = 100; // lunhezza primo link
float L2 = 70; // lunhezza secondo link  
float[] theta = new float[2];
float argCos2 = 0;
float C2 = 0;
float S2 = 0;
float theta1Des = 0;
float theta2Des = 0;
int nGiri = 0;
float kp = .3; // la kp del controllo
float precisione = .0001; // la differenza tra angolo finale e desiderato
float gomito = 1;

void setup() {
size(600, 400);
xBase = width/2;
yBase = height/2;
xCentro = xBase;
yCentro = yBase;
smooth();
theta[0] = 0;
theta[1] = PI/2;
}

void draw() 
{
  background(204);
  
  // Per modificare il kp e il gomito alto o basso
  if (keyPressed)
  {
    if (keyCode == UP)
    {
      kp += .001;
      if (kp > .5)
      {
        kp = .5;
      }
    }
    if (keyCode == DOWN)
    {
      kp -= .001;
      if (kp < 0.01)
      {
        kp = 0.01;
      }
    }
    if (keyCode == LEFT)
    {
      omega -= .005;
      if (omega < -1)
      {
        omega = -1;
      } 
      deltaPhi = (omegaOld-omega)*tempo + deltaPhi;
      omegaOld = omega;
      /* La necessità di assegnare una fase deltaPhi (e di aggiornarla) nell'argomento di seno e coseno (si veda sotto) 
      sta nel fatto che, cambiando la velocità di rotazione omega, non si vuole che cambi anche il punto desiderato.
      Quindi si assegna la fase deltaPhi in modo tale che omegaNuova*t+deltaPhiNuova = omegaOld*t + deltaPhiOld.
      Di qui si ha deltaPhiNuova = (omegaOld-omegaNuova)*t + deltaPhiOld che coincide con l'assegnazione di cui sopra */
    }
    if (keyCode == RIGHT)
    {
      omega += .005;
      if (omega > 1)
      {
        omega = 1;
      } 
      deltaPhi = (omegaOld-omega)*tempo + deltaPhi;
      omegaOld = omega;
    }
    if (key == '+')
    {
      rCerchio += 1;
      if (rCerchio > 200)
      {
        rCerchio = 200;
      }
    }
    if (key == '-')
    {
      rCerchio -= 1;
      if (rCerchio < 5)
      {
        rCerchio = 5;
      }
    }
    if ((key == 'a') || (key == 'A'))
    {
      gomito = -1;
    }
    if ((key == 'b') || (key == 'B'))
    {
      gomito = 1;
    }
  }    
  
  // inserisci coordinate effettore desiderate
  if (mousePressed)
  {
    xCentro = mouseX;
    yCentro = mouseY;
  }
  
  noFill();
  strokeWeight(2);
  stroke(0,0,255);
  ellipse(xCentro,yCentro,2*rCerchio,2*rCerchio);
  
  
  tempo = millis()/100; // tempo trascorso dall'inizio dell'esecuzione del programma
  
  
  xDes = xCentro + rCerchio*cos(omega*tempo+deltaPhi) - xBase;
  yDes = yBase - (yCentro - rCerchio*sin(omega*tempo+deltaPhi));

  // calcola coordinate giunto:
  argCos2 = (pow(xDes,2)+pow(yDes,2)-pow(L1,2)-pow(L2,2))/(2*L1*L2);
  if (abs(argCos2)<=1)
  {
    theta2Des = gomito*acos(argCos2);
    C2 = cos(theta2Des);
    S2 = sin(theta2Des);
    theta1Des = atan2(-L2*S2*xDes+(L1+L2*C2)*yDes,(L1+L2*C2)*xDes+L2*S2*yDes) + nGiri*2*PI;
    /* La quantità nGiri*2*PI viene aggiunta (e successivamente aggiornata come effettuato nelle righe che seguono) a theta1Des 
    per evitare che theta1Des passi istantaneamente da, per esempio, 179 gradi a -179 gradi. Questo accade perché l'atan2 
    restituisce angoli tra -180 e 180. Se theta1Des precedente valeva 179 gradi (e quindi anche theta1 del manipolatore è prossimo 
    a quel valore) e theta1Des attuale viene -179, quest'ultimo va visto come 181 piuttosto che come -179: in questo modo l'obiettivo
    di theta1 si muove solo di due gradi piuttosto che di 179-(-179)=358. */      
    
    if (abs(theta[0]-theta1Des) > abs(theta[0]-(theta1Des + 2*PI)))
    {
      theta1Des = theta1Des + 2*PI;
      nGiri += 1;
    }
    if (abs(theta[0]-theta1Des) > abs(theta[0]-(theta1Des - 2*PI)))
    {
      theta1Des = theta1Des - 2*PI;
      nGiri -= 1;
    }
    
    
    if (abs(theta[0]-theta1Des)>precisione)
    {
      theta[0] += kp*(theta1Des-theta[0]);
    }
    if (abs(theta[1]-theta2Des)>precisione)
    {
      theta[1] += kp*(theta2Des-theta[1]);
    }
  }
  else
  {
    textSize(25);
    fill(256,0,0);
    text("posizione fuori dello spazio di lavoro",width-500,height-20); 
  }
    
  robot(xBase,yBase,theta[0],theta[1]);
// Disegno punto obiettivo su cerchio
  noStroke();
  fill(0,0,255);
  ellipse(xBase+xDes,yBase-yDes,10,10);
  
  textSize(25);
  fill(0,0,256);
  text("kp = ",width-170,30); 
  text(kp,width-110,30);
  if (gomito == -1)
  {
    fill(0,256,0);
    text("Gomito alto",width-170,60); 
  }
  else
  {
    fill(0,256,0);
    text("Gomito basso",width-170,60); 
  }  
  textSize(15);
  fill(0);
  text("xDes = ",10,20); 
  text(xDes,70,20);
  text("yDes = ",150,20); 
  text(yDes,210,20);
  text("xEff = ",10,50); 
  text(L1*cos(theta[0])+L2*cos(theta[0]+theta[1]),70,50);
  text("yEff = ",150,50); 
  text((L1*sin(theta[0])+L2*sin(theta[0]+theta[1])),210,50);
  text("rCerchio = ",10,80); 
  text(rCerchio,100,80);
  text("omega = ",200,80); 
  text(omega,270,80);
  text("nGiri = ",10,110); 
  text(nGiri,70,110);
}

void robot(float x, float y, float theta1, float theta2)
{
  // Qui le coordinate si riferiscono allo schermo: (0,0) = in alto a sinistra
  // coordinate grafiche primo giunto
  float x1 = x+L1*cos(theta1);
  float y1 = y-L1*sin(theta1);  // Il meno per la convenzione grafica della y positiva verso il basso
  // coordinate grafiche secondo giunto  
  float x2 = x1+L2*cos(theta1+theta2);
  float y2 = y1-L2*sin(theta1+theta2);  
  
  noStroke();
  fill(0);
  // disegno base (nera)
  rect(x-15,y-15,30,30);
  
  fill(255,0,0);
  
  // disegno giunti (rossi)
  ellipse(x,y,15,15);
//  fill(255,0,0);
  ellipse(x1,y1,15,15);
  ellipse(x2,y2,15,15);

  stroke(0);
  strokeWeight(8);
  line(x,y,x1,y1);
  line(x1,y1,x2,y2);
  
//  float xPuntoP1 = x2-7*sin(theta1+theta2);
//  float yPuntoP1 = y2-7*cos(theta1+theta2);  
//  float xPuntoP2 = x2+7*sin(theta1+theta2);
//  float yPuntoP2 = y2+7*cos(theta1+theta2);  
//  float xPuntoP11 = xPuntoP1+8*cos(theta1+theta2);
//  float yPuntoP11 = yPuntoP1-8*sin(theta1+theta2);
//  float xPuntoP22 = xPuntoP2+8*cos(theta1+theta2);
//  float yPuntoP22 = yPuntoP2-8*sin(theta1+theta2);
//  line(xPuntoP1,yPuntoP1,xPuntoP2,yPuntoP2);
//  line(xPuntoP1,yPuntoP1,xPuntoP11,yPuntoP11);
//  line(xPuntoP2,yPuntoP2,xPuntoP22,yPuntoP22);
}